

#include <iostream>
#include <time.h>
#include <fstream>
#include <stdlib.h>

#define byte unsigned char
#define true 1
#define false 0
#define long int

#define DISPLAY_WIDTH 16

#define LEFT_BUTTON_MASK 0x80
#define RIGHT_BUTTON_MASK 0x40
#define UP_BUTTON_MASK 0x20
#define DOWN_BUTTON_MASK 0x10
#define RETURN_BUTTON_MASK 0x08
#define ESCAPE_BUTTON_MASK 0x04

#define MAXIMUM_FILE_NAME_LENGTH 16
#define FILE_DATA_OFFSET MAXIMUM_FILE_NAME_LENGTH + 1
#define FILE_ENTRY_SIZE 4096
#define NUMBER_OF_FILE_ENTRY_POSITIONS 32
#define EMPTY_FILE_ENTRY_INDICATOR 0xFF
#define FILE_BUFFER_SIZE 100

#define SPECIAL_CHARACTER_LENGTH 6
#define NUMBER_OF_SPECIAL_CHARACTERS 2
#define TOTAL_NUMBER_OF_CHARACTERS NUMBER_OF_SPECIAL_CHARACTERS + sizeof(CHARACTER_SET) - 1
#define SPACE_CHARACTER_INDEX 0
#define DELETE_CHARACTER_INDEX 1
#define EDIT_CURSOR_CHARACTER '_'

#define LITERAL_ARGUMENT_ADDRESS_LIST_OFFSET 0
#define LITERAL_ARGUMENT_ADDRESS_LIST_SIZE 20
#define STACK_OFFSET LITERAL_ARGUMENT_ADDRESS_LIST_OFFSET + LITERAL_ARGUMENT_ADDRESS_LIST_SIZE
#define SCOPE_RETURN_ADDRESS_OFFSET 0
#define SCOPE_PREVIOUS_SCOPE_ADDRESS_OFFSET 4
#define SCOPE_SIZE_OFFSET 6
#define SCOPE_VARIABLE_LIST_OFFSET 8
#define NUMBER_OF_SCOPE_VARIABLES 26
#define SCOPE_FLOW_DATA_OFFSET SCOPE_VARIABLE_LIST_OFFSET + NUMBER_OF_SCOPE_VARIABLES * 2

#define HEAP_START_ADDRESS 32000
#define HEAP_ENTRY_SIZE 8
#define HEAP_ENTRY_TYPE_OFFSET 0
#define HEAP_ENTRY_REFERENCE_COUNT_OFFSET 2
#define HEAP_ENTRY_DATA_OFFSET 4
#define HEAP_ENTRY_LINK_OFFSET 6

#define EMPTY_HEAP_ENTRY_TYPE 0
#define INTEGER_HEAP_ENTRY_TYPE 1
#define LIST_HEAP_ENTRY_TYPE 2

#define INTERPRET_FLOW_DATA -1
#define IGNORE_FLOW_DATA -2

const short RANDOM_DATA_LIST_1[] = {26300, 12613, 26904, 8022, 30794, 31703, 25650, 2068, 26336, 26781, 16264, 19980, 15295, 31750, 3123, 32465, 4086, 14700, 31978};
const short RANDOM_DATA_LIST_2[] = {29646, 3873, 6645, 27385, 11518, 9321, 2002, 31546, 5100, 12871, 15150, 10975, 23235, 16316, 10161, 745, 27271, 26236, 7635, 9953, 15108, 30539, 16157, 16197, 20820, 21735, 24581, 14531, 21504, 21949, 27284};

const byte CHARACTER_SET[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789,.?!;:'\"()=<>+-*/%~|&";
const byte SPECIAL_CHARACTER_SET[] = "SPACE DELETE";

const byte MESSAGE_1[] = "SAVED";
const byte MESSAGE_2[] = "FILE NAME?";
const byte MESSAGE_3[] = "CREATED";
const byte MESSAGE_4[] = "ARE YOU SURE?";
const byte MESSAGE_5[] = "DELETED";
const byte MESSAGE_6[] = "RENAMED";
const byte MESSAGE_7[] = "WELCOME";
const byte MESSAGE_8[] = "NO FILES";
const byte MESSAGE_9[] = "GO AWAY!";
const byte SELECTION_ITEM_1[] = "INSERT";
const byte SELECTION_ITEM_2[] = "DELETE";
const byte SELECTION_ITEM_3[] = "EDIT";
const byte SELECTION_ITEM_4[] = "SAVE";
const byte SELECTION_ITEM_5[] = "QUIT";
const byte SELECTION_ITEM_6[] = "CREATE";
const byte SELECTION_ITEM_7[] = "SELECT";
const byte SELECTION_ITEM_8[] = "EDIT";
const byte SELECTION_ITEM_9[] = "RUN";
const byte SELECTION_ITEM_10[] = "DELETE";
const byte SELECTION_ITEM_11[] = "RENAME";
const byte SELECTION_ITEM_12[] = "CANCEL";
const byte SELECTION_ITEM_13[] = "YES DELETE";
const byte * const SELECTION_MENU_1[] = {SELECTION_ITEM_1, SELECTION_ITEM_2, SELECTION_ITEM_3};
const byte * const SELECTION_MENU_2[] = {SELECTION_ITEM_4, SELECTION_ITEM_5};
const byte * const SELECTION_MENU_3[] = {SELECTION_ITEM_6, SELECTION_ITEM_7};
const byte * const SELECTION_MENU_4[] = {SELECTION_ITEM_8, SELECTION_ITEM_9, SELECTION_ITEM_10, SELECTION_ITEM_11};
const byte * const SELECTION_MENU_5[] = {SELECTION_ITEM_12, SELECTION_ITEM_13};

const byte BUILT_IN_FUNCTION_NAME_LIST[] = "= + - * / % == > ! ~ | & << >> IF END WHL BRK RET RAND STR INT LEN TRUNC GET SET PRINT INPUT ";

short randomNumber = 0;
byte randomNumberState1 = 0;
byte randomNumberState2 = 0;

long commandAddress;
short scopeAddress;
short heapSize = 0;
byte isIgnoringCommands;
short argumentPointerAddressList[10];
byte hasStoppedExecution;

byte displayBuffer[(DISPLAY_WIDTH + 1) * 2 + 1];
byte displayPosX = 0;
byte displayPosY = 0;

byte sramBuffer[33000];
byte eepromBuffer[128000];

byte pgm_read_byte(const byte *pointer)
{
    return *pointer;
}

short pgm_read_word(const short *pointer)
{
    return *pointer;
}

// I wrote this because rand takes up more room.
// The RNG does not need to be extremely robust.
short generateRandomNumber()
{
    randomNumber <<= 1;
    if (randomNumber & 0x8000)
    {
        randomNumber &= 0x7FFF;
        randomNumber |= 0x0001;
    }
    randomNumber ^= pgm_read_word(RANDOM_DATA_LIST_1 + randomNumberState1);
    randomNumber ^= pgm_read_word(RANDOM_DATA_LIST_2 + randomNumberState2);
    randomNumberState1 += 1;
    randomNumberState2 += 1;
    if (randomNumberState1 >= sizeof(RANDOM_DATA_LIST_1))
    {
        randomNumberState1 = 0;
    }
    if (randomNumberState2 >= sizeof(RANDOM_DATA_LIST_2))
    {
        randomNumberState2 = 0;
    }
    return randomNumber;
}

short getTextLength(byte *text)
{
    short output = 0;
    while (true)
    {
        byte tempCharacter = *text;
        if (tempCharacter == 0)
        {
            return output;
        }
        text += 1;
        output += 1;
    }
}

byte equalText(byte *text1, byte *text2)
{
    while (true)
    {
        byte tempCharacter1 = *text1;
        byte tempCharacter2 = *text2;
        if (tempCharacter1 != tempCharacter2)
        {
            return false;
        }
        if (tempCharacter1 == 0)
        {
            return true;
        }
        text1 += 1;
        text2 += 1;
    }
}

void sendDisplayCharacter(byte character)
{
    if (displayPosX >= DISPLAY_WIDTH)
    {
        return;
    }
    byte index = displayPosX + displayPosY * (DISPLAY_WIDTH + 1);
    displayBuffer[index] = character;
    displayPosX += 1;
}

void clearDisplay()
{
    byte index = 0;
    byte tempPosY = 0;
    while (tempPosY < 2)
    {
        byte tempPosX = 0;
        while (tempPosX < DISPLAY_WIDTH)
        {
            displayBuffer[index] = ' ';
            index += 1;
            tempPosX += 1;
        }
        displayBuffer[index] = '\n';
        index += 1;
        tempPosY += 1;
    }
}

void setDisplayPos(byte posX, byte posY)
{
    displayPosX = posX;
    displayPosY = posY;
}

void clearDisplayRow(byte posY)
{
    setDisplayPos(0, posY);
    byte tempCount = 0;
    while (tempCount < DISPLAY_WIDTH)
    {
        sendDisplayCharacter(' ');
        tempCount += 1;
    }
}

void readSramData(byte *data, short amount, short address)
{
    short tempCount = 0;
    while (tempCount < amount)
    {
        *data = sramBuffer[address];
        address += 1;
        data += 1;
        tempCount += 1;
    }
}

void writeSramData(short address, byte *data, short amount)
{
    short tempCount = 0;
    while (tempCount < amount)
    {
        sramBuffer[address] = *data;
        address += 1;
        data += 1;
        tempCount += 1;
    }
}

byte readSramByte(short address)
{
    byte output;
    readSramData(&output, 1, address);
    return output;
}

short readSramShort(short address)
{
    short output;
    readSramData((byte *)&output, 2, address);
    return output;
}

long readSramLong(short address)
{
    long output;
    readSramData((byte *)&output, 4, address);
    return output;
}

void writeSramByte(short address, byte value)
{
    writeSramData(address, &value, 1);
}

void writeSramShort(short address, short value)
{
    writeSramData(address, (byte *)&value, 2);
}

void writeSramLong(short address, long value)
{
    writeSramData(address, (byte *)&value, 4);
}

void readEepromData(byte *data, short amount, long address)
{
    short tempCount = 0;
    while (tempCount < amount)
    {
        *data = eepromBuffer[address];
        address += 1;
        data += 1;
        tempCount += 1;
    }
}

// Note: Page write only works within 256 byte boundaries.
void writeEepromPage(long address, byte *data, short amount)
{
    short tempCount = 0;
    while (tempCount < amount)
    {
        eepromBuffer[address] = *data;
        address += 1;
        data += 1;
        tempCount += 1;
    }
}

void writeEepromData(long address, byte *data, short amount)
{
    short tempCount = 0;
    while (tempCount < amount)
    {
        long tempEndAddress = (address + 256) & 0xFFFFFF00;
        short tempAmount = tempEndAddress - address;
        short tempAmount2 = amount - tempCount;
        if (tempAmount > tempAmount2)
        {
            tempAmount = tempAmount2;
        }
        writeEepromPage(address, data, tempAmount);
        address = tempEndAddress;
        data += tempAmount;
        tempCount += tempAmount;
    }
}

byte readEepromByte(long address)
{
    byte output;
    readEepromData(&output, 1, address);
    return output;
}

// Destination should have size at least MAXIMUM_FILE_NAME_LENGTH + 1.
void getFileName(byte *destination, byte index)
{
    long tempAddress = index * (long)FILE_ENTRY_SIZE;
    readEepromData(destination, MAXIMUM_FILE_NAME_LENGTH + 1, tempAddress);
}

void loadFile(byte index)
{
    long tempStartAddress = index * (long)FILE_ENTRY_SIZE + FILE_DATA_OFFSET;
    short tempOffset = 0;
    short tempEndOffset = FILE_ENTRY_SIZE - FILE_DATA_OFFSET;
    while (tempOffset < tempEndOffset)
    {
        short tempAmount = FILE_BUFFER_SIZE;
        if (tempOffset + tempAmount > tempEndOffset)
        {
            tempAmount = tempEndOffset - tempOffset;
        }
        byte tempBuffer[tempAmount];
        readEepromData(tempBuffer, tempAmount, tempStartAddress + tempOffset);
        writeSramData(tempOffset, tempBuffer, tempAmount);
        tempOffset += tempAmount;
    }
}

void saveFile(byte index)
{
    long tempStartAddress = index * (long)FILE_ENTRY_SIZE + FILE_DATA_OFFSET;
    short tempOffset = 0;
    short tempEndOffset = FILE_ENTRY_SIZE - FILE_DATA_OFFSET;
    while (tempOffset < tempEndOffset)
    {
        short tempAmount = FILE_BUFFER_SIZE;
        if (tempOffset + tempAmount > tempEndOffset)
        {
            tempAmount = tempEndOffset - tempOffset;
        }
        byte tempBuffer[tempAmount];
        readSramData(tempBuffer, tempAmount, tempOffset);
        writeEepromData(tempStartAddress + tempOffset, tempBuffer, tempAmount);
        tempOffset += tempAmount;
    }
}

byte promptButton()
{
    byte output = 0;
    while (!output)
    {
        std::cout << std::endl << std::endl << std::endl << std::endl << std::endl << std::endl << std::endl << std::endl << displayBuffer;
        byte tempBuffer[20];
        std::cin >> tempBuffer;
        byte tempCharacter = tempBuffer[0];
        if (tempCharacter == 'a')
        {
            output |= LEFT_BUTTON_MASK;
        }
        if (tempCharacter == 'd')
        {
            output |= RIGHT_BUTTON_MASK;
        }
        if (tempCharacter == 'w')
        {
            output |= UP_BUTTON_MASK;
        }
        if (tempCharacter == 's')
        {
            output |= DOWN_BUTTON_MASK;
        }
        if (tempCharacter == 'e')
        {
            output |= RETURN_BUTTON_MASK;
        }
        if (tempCharacter == 'q')
        {
            output |= ESCAPE_BUTTON_MASK;
        }
    }
    return output;
}

void displayText(byte *text)
{
    clearDisplay();
    setDisplayPos(0, 0);
    byte index = 0;
    while (true)
    {
        byte tempCharacter = text[index];
        if (tempCharacter == 0)
        {
            break;
        }
        if (index == DISPLAY_WIDTH)
        {
            setDisplayPos(0, 1);
        }
        sendDisplayCharacter(tempCharacter);
        index += 1;
    }
}

void displayProgmemText(const byte *text)
{
    clearDisplay();
    setDisplayPos(0, 0);
    byte index = 0;
    while (true)
    {
        byte tempCharacter = pgm_read_byte(text + index);
        if (tempCharacter == 0)
        {
            break;
        }
        if (index == DISPLAY_WIDTH)
        {
            setDisplayPos(0, 1);
        }
        sendDisplayCharacter(tempCharacter);
        index += 1;
    }
}

// Returns 255 if escape was pressed.
byte promptProgmemSelection(const byte * const *textList, byte amount)
{
    byte output = 0;
    while (true)
    {
        displayProgmemText(textList[output]);
        byte tempButtons = promptButton();
        if (tempButtons & (LEFT_BUTTON_MASK | UP_BUTTON_MASK))
        {
            if (output)
            {
                output -= 1;
            } else {
                output = amount - 1;
            }
        }
        if (tempButtons & (RIGHT_BUTTON_MASK | DOWN_BUTTON_MASK))
        {
            output += 1;
            if (output >= amount)
            {
                output = 0;
            }
        }
        if (tempButtons & RETURN_BUTTON_MASK)
        {
            return output;
        }
        if (tempButtons & ESCAPE_BUTTON_MASK)
        {
            return 255;
        }
    }
}

void editTextLine(byte *text)
{
    // 0 = Navigate.
    // 1 = Edit.
    byte tempEditState = 0;
    short tempCursorIndex = 0;
    short tempCharacterIndex = 0;
    while (true)
    {
        clearDisplay();
        setDisplayPos(0, 0);
        short tempLength = getTextLength(text);
        short tempStartIndex;
        if (tempCursorIndex < DISPLAY_WIDTH / 2)
        {
            tempStartIndex = 0;
        } else
        {
            tempStartIndex = tempCursorIndex - DISPLAY_WIDTH / 2;
        }
        byte hasFoundLastCharacter = false;
        byte tempCount = 0;
        byte tempOffset = 0;
        while (tempOffset < DISPLAY_WIDTH)
        {
            short tempIndex = tempStartIndex + tempOffset;
            if (tempIndex == tempCursorIndex)
            {
                sendDisplayCharacter(EDIT_CURSOR_CHARACTER);
                tempCount += 1;
            }
            if (tempCount < DISPLAY_WIDTH)
            {
                byte tempCharacter = text[tempIndex];
                if (tempCharacter == 0)
                {
                    hasFoundLastCharacter = true;
                }
                if (!hasFoundLastCharacter)
                {
                    sendDisplayCharacter(tempCharacter);
                    tempCount += 1;
                }
            }
            tempOffset += 1;
        }
        if (tempEditState == 0)
        {
            byte tempButtons = promptButton();
            if (tempButtons & LEFT_BUTTON_MASK)
            {
                tempCursorIndex -= 1;
            }
            if (tempButtons & RIGHT_BUTTON_MASK)
            {
                tempCursorIndex += 1;
            }
            if (tempButtons & UP_BUTTON_MASK)
            {
                tempCursorIndex -= 8;
            }
            if (tempButtons & DOWN_BUTTON_MASK)
            {
                tempCursorIndex += 8;
            }
            if (tempCursorIndex < 0)
            {
                tempCursorIndex = 0;
            }
            if (tempCursorIndex > tempLength)
            {
                tempCursorIndex = tempLength;
            }
            if (tempButtons & RETURN_BUTTON_MASK)
            {
                tempEditState = 1;
            }
            if (tempButtons & ESCAPE_BUTTON_MASK)
            {
                return;
            }
        } else if (tempEditState == 1)
        {
            while (true)
            {
                clearDisplayRow(1);
                setDisplayPos(0, 1);
                if (tempCharacterIndex < NUMBER_OF_SPECIAL_CHARACTERS)
                {
                    byte tempOffset = 0;
                    while (tempOffset < SPECIAL_CHARACTER_LENGTH)
                    {
                        byte tempCharacter = pgm_read_byte(SPECIAL_CHARACTER_SET + tempCharacterIndex * SPECIAL_CHARACTER_LENGTH + tempOffset);
                        sendDisplayCharacter(tempCharacter);
                        tempOffset += 1;
                    }
                } else {
                    byte tempCharacter = pgm_read_byte(CHARACTER_SET + tempCharacterIndex - NUMBER_OF_SPECIAL_CHARACTERS);
                    sendDisplayCharacter(tempCharacter);
                }
                byte tempButtons = promptButton();
                if (tempButtons & LEFT_BUTTON_MASK)
                {
                    tempCharacterIndex -= 1;
                }
                if (tempButtons & RIGHT_BUTTON_MASK)
                {
                    tempCharacterIndex += 1;
                }
                if (tempButtons & UP_BUTTON_MASK)
                {
                    tempCharacterIndex -= 8;
                }
                if (tempButtons & DOWN_BUTTON_MASK)
                {
                    tempCharacterIndex += 8;
                }
                if (tempCharacterIndex < 0)
                {
                    tempCharacterIndex += TOTAL_NUMBER_OF_CHARACTERS;
                }
                if (tempCharacterIndex >= TOTAL_NUMBER_OF_CHARACTERS)
                {
                    tempCharacterIndex -= TOTAL_NUMBER_OF_CHARACTERS;
                }
                if (tempButtons & RETURN_BUTTON_MASK)
                {
                    if (tempCharacterIndex == DELETE_CHARACTER_INDEX) {
                        if (tempCursorIndex > 0)
                        {
                            short index = tempCursorIndex;
                            while (index <= tempLength)
                            {
                                byte tempCharacter = text[index];
                                text[index - 1] = tempCharacter;
                                index += 1;
                            }
                            tempCursorIndex -= 1;
                        }
                    } else {
                        byte tempCharacter;
                        if (tempCharacterIndex == SPACE_CHARACTER_INDEX)
                        {
                            tempCharacter = ' ';
                        } else {
                            tempCharacter = pgm_read_byte(CHARACTER_SET + tempCharacterIndex - NUMBER_OF_SPECIAL_CHARACTERS);
                        }
                        short index = tempLength;
                        while (index >= tempCursorIndex)
                        {
                            byte tempCharacter2 = text[index];
                            text[index + 1] = tempCharacter2;
                            index -= 1;
                        }
                        text[tempCursorIndex] = tempCharacter;
                        tempCursorIndex += 1;
                    }
                    break;
                }
                if (tempButtons & ESCAPE_BUTTON_MASK)
                {
                    tempEditState = 0;
                    break;
                }
            }
        }
    }
}

short findPreviousLine(short lineIndex)
{
    lineIndex -= 2;
    while (lineIndex >= 0)
    {
        byte tempCharacter = readSramByte(lineIndex);
        if (tempCharacter == '\n')
        {
            return lineIndex + 1;
        }
        lineIndex -= 1;
    }
    return 0;
}

short findNextLine(short lineIndex)
{
    short startLineIndex = lineIndex;
    while (true)
    {
        byte tempCharacter = readSramByte(lineIndex);
        if (tempCharacter == '\n')
        {
            return lineIndex + 1;
        }
        if (tempCharacter == 0)
        {
            return startLineIndex;
        }
        lineIndex += 1;
    }
}

void moveLines(short startIndex, short endIndex)
{
    short tempOffset = 0;
    if (endIndex > startIndex)
    {
        while (true)
        {
            byte tempCharacter = readSramByte(startIndex + tempOffset);
            if (tempCharacter == 0)
            {
                break;
            }
            tempOffset += 1;
        }
        while (tempOffset >= 0)
        {
            byte tempCharacter = readSramByte(startIndex + tempOffset);
            writeSramByte(endIndex + tempOffset, tempCharacter);
            tempOffset -= 1;
        }
    }
    if (endIndex < startIndex)
    {
        while (true)
        {
            byte tempCharacter = readSramByte(startIndex + tempOffset);
            writeSramByte(endIndex + tempOffset, tempCharacter);
            if (tempCharacter == 0)
            {
                break;
            }
            tempOffset += 1;
        }
    }
}

void editFile(byte fileIndex)
{
    loadFile(fileIndex);
    short tempLineIndex = 0;
    while (true)
    {
        clearDisplay();
        setDisplayPos(0, 0);
        byte tempOffset = 0;
        while (tempOffset < DISPLAY_WIDTH * 2)
        {
            byte tempCharacter = readSramByte(tempLineIndex + tempOffset);
            if (tempCharacter == '\n' || tempCharacter == 0)
            {
                break;
            }
            if (tempOffset == DISPLAY_WIDTH)
            {
                setDisplayPos(0, 1);
            }
            sendDisplayCharacter(tempCharacter);
            tempOffset += 1;
        }
        byte tempButtons = promptButton();
        if (tempButtons & LEFT_BUTTON_MASK)
        {
            tempLineIndex = findPreviousLine(tempLineIndex);
        }
        if (tempButtons & RIGHT_BUTTON_MASK)
        {
            tempLineIndex = findNextLine(tempLineIndex);
        }
        if (tempButtons & UP_BUTTON_MASK)
        {
            int tempCount = 0;
            while (tempCount < 8)
            {
                tempLineIndex = findPreviousLine(tempLineIndex);
                tempCount += 1;
            }
        }
        if (tempButtons & DOWN_BUTTON_MASK)
        {
            int tempCount = 0;
            while (tempCount < 8)
            {
                tempLineIndex = findNextLine(tempLineIndex);
                tempCount += 1;
            }
        }
        if (tempButtons & RETURN_BUTTON_MASK)
        {
            byte tempResult = promptProgmemSelection(SELECTION_MENU_1, 3);
            // Insert.
            if (tempResult == 0)
            {
                byte tempBuffer[256];
                tempBuffer[0] = 0;
                editTextLine(tempBuffer);
                byte tempLength = getTextLength(tempBuffer);
                tempBuffer[tempLength] = '\n';
                tempLength += 1;
                moveLines(tempLineIndex, tempLineIndex + tempLength);
                writeSramData(tempLineIndex, tempBuffer, tempLength);
                tempLineIndex = findNextLine(tempLineIndex);
            }
            // Delete.
            if (tempResult == 1)
            {
                short tempNextLineIndex = findNextLine(tempLineIndex);
                moveLines(tempNextLineIndex, tempLineIndex);
                tempLineIndex = findPreviousLine(tempLineIndex);
            }
            // Edit.
            if (tempResult == 2)
            {
                byte tempBuffer[256];
                byte tempOffset = 0;
                byte tempLength = 0;
                while (true)
                {
                    byte tempCharacter = readSramByte(tempLineIndex + tempOffset);
                    if (tempCharacter == 0)
                    {
                        break;
                    } else {
                        tempLength += 1;
                    }
                    if (tempCharacter == '\n')
                    {
                        break;
                    }
                    tempBuffer[tempOffset] = tempCharacter;
                    tempOffset += 1;
                }
                tempBuffer[tempOffset] = 0;
                editTextLine(tempBuffer);
                byte tempLength2 = getTextLength(tempBuffer);
                tempBuffer[tempLength2] = '\n';
                tempLength2 += 1;
                moveLines(tempLineIndex + tempLength, tempLineIndex + tempLength2);
                writeSramData(tempLineIndex, tempBuffer, tempLength2);
            }
        }
        if (tempButtons & ESCAPE_BUTTON_MASK)
        {
            byte tempResult = promptProgmemSelection(SELECTION_MENU_2, 2);
            // Save.
            if (tempResult == 0)
            {
                saveFile(fileIndex);
                displayProgmemText(MESSAGE_1);
                promptButton();
            }
            // Quit.
            if (tempResult == 1)
            {
                return;
            }
        }
    }
}

byte getNumberOfFiles()
{
    byte output = 0;
    long tempAddress = 0;
    byte index = 0;
    while (index < NUMBER_OF_FILE_ENTRY_POSITIONS)
    {
        byte tempValue = readEepromByte(tempAddress);
        if (tempValue != EMPTY_FILE_ENTRY_INDICATOR)
        {
            output += 1;
        }
        tempAddress += FILE_ENTRY_SIZE;
        index += 1;
    }
    return output;
}

byte getFileIndexByNumber(byte number)
{
    byte output = 0;
    long tempAddress = 0;
    byte index = 0;
    while (index < NUMBER_OF_FILE_ENTRY_POSITIONS)
    {
        byte tempValue = readEepromByte(tempAddress);
        if (tempValue != EMPTY_FILE_ENTRY_INDICATOR)
        {
            if (number < 1)
            {
                return output;
            }
            number -= 1;
        }
        tempAddress += FILE_ENTRY_SIZE;
        output += 1;
        index += 1;
    }
    return 255;
}

void initializeScopeVariables()
{
    byte index = 0;
    while (index < NUMBER_OF_SCOPE_VARIABLES)
    {
        short tempAddress = scopeAddress + SCOPE_VARIABLE_LIST_OFFSET + index * 2;
        writeSramShort(tempAddress, 0);
        index += 1;
    }
}

// Returns 255 if the function could not be found.
byte findBuiltInFunction(byte *name)
{
    byte output = 0;
    short index = 0;
    while (index < sizeof(BUILT_IN_FUNCTION_NAME_LIST))
    {
        byte hasFoundDifference = false;
        byte tempOffset = 0;
        while (true)
        {
            byte tempCharacter = pgm_read_byte(BUILT_IN_FUNCTION_NAME_LIST + index);
            index += 1;
            if (!hasFoundDifference)
            {
                byte tempCharacter2 = *(name + tempOffset);
                tempOffset += 1;
                if (tempCharacter == ' ')
                {
                    if (tempCharacter2 == 0)
                    {
                        return output;
                    }
                    break;
                } else {
                    if (tempCharacter != tempCharacter2)
                    {
                        hasFoundDifference = true;
                    }
                }
            } else {
                if (tempCharacter == ' ')
                {
                    break;
                }
            }
        }
        output += 1;
    }
    return 255;
}

void setHeapEntryReference(short referenceAddress, short reference);

short getHeapEntryType(short address)
{
    return readSramShort(address + HEAP_ENTRY_TYPE_OFFSET);
}

short getHeapEntryData(short address)
{
    return readSramShort(address + HEAP_ENTRY_DATA_OFFSET);
}

short getHeapEntryLink(short address)
{
    return readSramShort(address + HEAP_ENTRY_LINK_OFFSET);
}

void setHeapEntryType(short address, short type)
{
    writeSramShort(address + HEAP_ENTRY_TYPE_OFFSET, type);
}

void setHeapEntryData(short address, short data)
{
    short tempAddress = address + HEAP_ENTRY_DATA_OFFSET;
    if (getHeapEntryType(address) == LIST_HEAP_ENTRY_TYPE)
    {
        setHeapEntryReference(tempAddress, data);
    } else {
        writeSramShort(tempAddress, data);
    }
}

void setHeapEntryLink(short address, short link)
{
    setHeapEntryReference(address + HEAP_ENTRY_LINK_OFFSET, link);
}

void changeHeapEntryReferenceCount(short address, short offset)
{
    short tempCount = readSramShort(address + HEAP_ENTRY_REFERENCE_COUNT_OFFSET);
    tempCount += offset;
    if (tempCount < 1)
    {
        short tempType = getHeapEntryType(address);
        if (tempType == LIST_HEAP_ENTRY_TYPE)
        {
            // All Heap entries in the list chain should have a reference count of 1.
            // We use a loop instead of a recursion to avoid memory overhead.
            while (true)
            {
                short tempNextAddress = getHeapEntryLink(address);
                // Using setHeapEntryLink would cause recursion.
                setHeapEntryData(address, 0);
                setHeapEntryType(address, EMPTY_HEAP_ENTRY_TYPE);
                if (!tempNextAddress)
                {
                    break;
                }
                address = tempNextAddress;
            }
        } else {
            setHeapEntryType(address, EMPTY_HEAP_ENTRY_TYPE);
        }
    } else {
        writeSramShort(address + HEAP_ENTRY_REFERENCE_COUNT_OFFSET, tempCount);
    }
}

short allocateHeapEntry(short type)
{
    short tempOffset = 0;
    byte hasFoundEmptyEntry = false;
    while (tempOffset < heapSize)
    {
        short tempType = readSramShort(HEAP_START_ADDRESS - tempOffset);
        if (tempType == EMPTY_HEAP_ENTRY_TYPE)
        {
            hasFoundEmptyEntry = true;
            break;
        }
        tempOffset += HEAP_ENTRY_SIZE;
    }
    if (!hasFoundEmptyEntry)
    {
        heapSize += HEAP_ENTRY_SIZE;
    }
    short tempAddress = HEAP_START_ADDRESS - tempOffset;
    writeSramShort(tempAddress + HEAP_ENTRY_TYPE_OFFSET, type);
    writeSramShort(tempAddress + HEAP_ENTRY_REFERENCE_COUNT_OFFSET, 0);
    writeSramShort(tempAddress + HEAP_ENTRY_DATA_OFFSET, 0);
    writeSramShort(tempAddress + HEAP_ENTRY_LINK_OFFSET, 0);
    return tempAddress;
}

short allocateInteger(short value)
{
    short output = allocateHeapEntry(INTEGER_HEAP_ENTRY_TYPE);
    setHeapEntryData(output, value);
    return output;
}

short allocateList(short pointer)
{
    short output = allocateHeapEntry(LIST_HEAP_ENTRY_TYPE);
    setHeapEntryData(output, pointer);
    return output;
}

short allocateText(byte *text)
{
    short output = 0;
    short tempPointer = 0;
    while (true)
    {
        byte tempCharacter = *text;
        short tempPointer2 = allocateInteger(tempCharacter);
        short tempPointer3 = allocateList(tempPointer2);
        if (tempPointer == 0)
        {
            output = tempPointer3;
        } else {
            setHeapEntryLink(tempPointer, tempPointer3);
        }
        tempPointer = tempPointer3;
        if (tempCharacter == 0)
        {
            return output;
        }
        text += 1;
    }
}

void getTextFromHeapEntry(byte *destination, short pointer)
{
    while (true)
    {
        short tempPointer = getHeapEntryData(pointer);
        byte tempCharacter = getHeapEntryData(tempPointer);
        *destination = tempCharacter;
        if (tempCharacter == 0)
        {
            break;
        }
        pointer = getHeapEntryLink(pointer);
        destination += 1;
    }
}

// Use this function for automatic heap maintainence.
void setHeapEntryReference(short referenceAddress, short reference)
{
    short tempOldReference = readSramShort(referenceAddress);
    if (tempOldReference != EMPTY_HEAP_ENTRY_TYPE)
    {
        changeHeapEntryReferenceCount(tempOldReference, -1);
    }
    writeSramShort(referenceAddress, reference);
    if (reference != EMPTY_HEAP_ENTRY_TYPE)
    {
        changeHeapEntryReferenceCount(reference, 1);
    }
}

short convertEepromTextToInt(long address, short *tempOffset)
{
    byte tempBuffer[20];
    byte index = 0;
    while (true)
    {
        byte tempCharacter = readEepromByte(address + *tempOffset);
        if ((tempCharacter < '0' || tempCharacter > '9') && tempCharacter != '-')
        {
            tempBuffer[index] = 0;
            break;
        }
        tempBuffer[index] = tempCharacter;
        *tempOffset += 1;
        index += 1;
    }
    return atoi((char *)tempBuffer);
}

// Returns a new commandOffset.
short parseArgumentTerm(short commandOffset, byte argumentIndex)
{
    byte tempCharacter = readEepromByte(commandAddress + commandOffset);
    if ((tempCharacter >= '0' && tempCharacter <= '9') || tempCharacter == '-')
    {
        short tempNumber = convertEepromTextToInt(commandAddress, &commandOffset);
        short tempPointer = allocateInteger(tempNumber);
        short tempPointerAddress = LITERAL_ARGUMENT_ADDRESS_LIST_OFFSET + argumentIndex * 2;
        setHeapEntryReference(tempPointerAddress, tempPointer);
        argumentPointerAddressList[argumentIndex] = tempPointerAddress;
    }
    if (tempCharacter >= 'A' && tempCharacter <= 'Z')
    {
        commandOffset += 1;
        short tempPointerAddress = scopeAddress + SCOPE_VARIABLE_LIST_OFFSET + (tempCharacter - 'A') * 2;
        argumentPointerAddressList[argumentIndex] = tempPointerAddress;
    }
    if (tempCharacter == '(')
    {
        short tempStartPointer = 0;
        short tempPointer = 0;
        commandOffset += 1;
        while (true)
        {
            byte tempCharacter = readEepromByte(commandAddress + commandOffset);
            if (tempCharacter == ')')
            {
                break;
            }
            if (tempCharacter == ' ')
            {
                commandOffset += 1;
            }
            commandOffset = parseArgumentTerm(commandOffset, argumentIndex);
            short tempPointerAddress = argumentPointerAddressList[argumentIndex];
            short tempPointer2 = readSramShort(tempPointerAddress);
            short tempPointer3 = allocateList(tempPointer2);
            if (tempPointer == 0)
            {
                tempStartPointer = tempPointer3;
            } else {
                setHeapEntryLink(tempPointer, tempPointer3);
            }
            tempPointer = tempPointer3;
        }
        if (tempStartPointer == 0)
        {
            tempStartPointer = allocateList(0);
        }
        short tempPointerAddress = LITERAL_ARGUMENT_ADDRESS_LIST_OFFSET + argumentIndex * 2;
        setHeapEntryReference(tempPointerAddress, tempStartPointer);
        argumentPointerAddressList[argumentIndex] = tempPointerAddress;
    }
    if (tempCharacter == '"')
    {
        short tempStartPointer = 0;
        short tempPointer = 0;
        commandOffset += 1;
        while (true)
        {
            byte tempCharacter = readEepromByte(commandAddress + commandOffset);
            commandOffset += 1;
            if (tempCharacter == '"')
            {
                tempCharacter = 0;
            }
            short tempPointer2 = allocateInteger(tempCharacter);
            short tempPointer3 = allocateList(tempPointer2);
            if (tempPointer == 0)
            {
                tempStartPointer = tempPointer3;
            } else {
                setHeapEntryLink(tempPointer, tempPointer3);
            }
            tempPointer = tempPointer3;
            if (tempCharacter == 0)
            {
                break;
            }
        }
        short tempPointerAddress = LITERAL_ARGUMENT_ADDRESS_LIST_OFFSET + argumentIndex * 2;
        setHeapEntryReference(tempPointerAddress, tempStartPointer);
        argumentPointerAddressList[argumentIndex] = tempPointerAddress;
    }
    return commandOffset;
}

short getArgumentPointer(byte index)
{
    return readSramShort(argumentPointerAddressList[index]);
}

short getArgumentInteger(byte index)
{
    return getHeapEntryData(getArgumentPointer(index));
}

short getFlowDataAddress()
{
    return scopeAddress + readSramShort(scopeAddress + SCOPE_SIZE_OFFSET) - 4;
}

void changeFlowDataAddress(short offset)
{
    short tempSize = readSramShort(scopeAddress + SCOPE_SIZE_OFFSET);
    tempSize += offset;
    writeSramShort(scopeAddress + SCOPE_SIZE_OFFSET, tempSize);
}

void executeNextCommand()
{
    long tempNextCommandAddress;
    short tempOffset = 0;
    while (tempOffset < LITERAL_ARGUMENT_ADDRESS_LIST_SIZE)
    {
        short tempAddress = LITERAL_ARGUMENT_ADDRESS_LIST_OFFSET + tempOffset;
        writeSramShort(tempAddress, 0);
        tempOffset += 2;
    }
    byte shouldQuitFile = false;
    byte tempCommandName[30];
    tempOffset = 0;
    while (true)
    {
        byte tempCharacter = readEepromByte(commandAddress + tempOffset);
        if (tempCharacter == 0)
        {
            shouldQuitFile = true;
            break;
        }
        if (tempCharacter == ' ' || tempCharacter == '\n')
        {
            break;
        }
        tempCommandName[tempOffset] = tempCharacter;
        tempOffset += 1;
    }
    tempCommandName[tempOffset] = 0;
    byte tempCommand = findBuiltInFunction(tempCommandName);
    if (isIgnoringCommands)
    {
        while (true)
        {
            byte tempCharacter = readEepromByte(commandAddress + tempOffset);
            if (tempCharacter == 0)
            {
                shouldQuitFile = true;
                break;
            }
            if (tempCharacter == ' ' || tempCharacter == '\n')
            {
                break;
            }
            tempCommandName[tempOffset] = tempCharacter;
            tempOffset += 1;
        }
        tempNextCommandAddress = commandAddress + tempOffset + 1;
        switch(tempCommand)
        {
                // IF.
            case 14:
                // WHL.
            case 16:
            {
                changeFlowDataAddress(4);
                short tempAddress = getFlowDataAddress();
                writeSramLong(tempAddress, IGNORE_FLOW_DATA);
                break;
            }
                // END.
            case 15:
            {
                short tempAddress = getFlowDataAddress();
                long tempFlowData = readSramLong(tempAddress);
                changeFlowDataAddress(-4);
                if (tempFlowData == -1)
                {
                    isIgnoringCommands = false;
                }
                break;
            }
            default:
            {
                break;
            }
        }
    } else {
        byte tempArgumentIndex = 0;
        while (true)
        {
            byte tempCharacter = readEepromByte(commandAddress + tempOffset);
            if (tempCharacter != ' ')
            {
                break;
            }
            tempOffset = parseArgumentTerm(tempOffset + 1, tempArgumentIndex);
            tempArgumentIndex += 1;
        }
        byte tempNumberOfArguments = tempArgumentIndex;
        tempNextCommandAddress = commandAddress + tempOffset + 1;
        short tempValue1;
        short tempValue2;
        if (tempNumberOfArguments > 1)
        {
            tempValue1 = getArgumentInteger(1);
            if (tempNumberOfArguments > 2)
            {
                tempValue2 = getArgumentInteger(2);
            }
        }
        switch(tempCommand)
        {
                // =.
            case 0:
            {
                setHeapEntryReference(argumentPointerAddressList[0], getArgumentPointer(1));
                break;
            }
                // +.
            case 1:
            {
                short tempResult = tempValue1 + tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // -.
            case 2:
            {
                short tempResult = tempValue1 - tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // *.
            case 3:
            {
                short tempResult = tempValue1 * tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // /.
            case 4:
            {
                short tempResult = tempValue1 / tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // %.
            case 5:
            {
                short tempResult = tempValue1 % tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // ==.
            case 6:
            {
                short tempResult = tempValue1 == tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // >.
            case 7:
            {
                short tempResult = tempValue1 > tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // !.
            case 8:
            {
                short tempResult = !tempValue1;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // ~.
            case 9:
            {
                short tempResult = ~tempValue1;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // |.
            case 10:
            {
                short tempResult = tempValue1 | tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // &.
            case 11:
            {
                short tempResult = tempValue1 & tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // <<.
            case 12:
            {
                short tempResult = tempValue1 << tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // >>.
            case 13:
            {
                short tempResult = tempValue1 >> tempValue2;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // IF.
            case 14:
            {
                short tempValue = getArgumentInteger(0);
                if (!tempValue)
                {
                    isIgnoringCommands = true;
                }
                changeFlowDataAddress(4);
                short tempAddress = getFlowDataAddress();
                writeSramLong(tempAddress, INTERPRET_FLOW_DATA);
                break;
            }
                // END.
            case 15:
            {
                short tempAddress = getFlowDataAddress();
                long tempFlowData = readSramLong(tempAddress);
                changeFlowDataAddress(-4);
                if (tempFlowData >= 0)
                {
                    tempNextCommandAddress = tempFlowData;
                    shouldQuitFile = false;
                }
                break;
            }
                // WHL.
            case 16:
            {
                long tempFlowData;
                short tempValue = getArgumentInteger(0);
                if (tempValue)
                {
                    tempFlowData = commandAddress;
                } else {
                    tempFlowData = INTERPRET_FLOW_DATA;
                    isIgnoringCommands = true;
                }
                changeFlowDataAddress(4);
                short tempAddress = getFlowDataAddress();
                writeSramLong(tempAddress, tempFlowData);
                break;
            }
                // BRK.
            case 17:
            {
                isIgnoringCommands = true;
                short tempAddress = getFlowDataAddress();
                while (true)
                {
                    long tempFlowData = readSramLong(tempAddress);
                    if (tempFlowData == -1)
                    {
                        writeSramLong(tempAddress, -2);
                    }
                    if (tempFlowData >= 0)
                    {
                        writeSramLong(tempAddress, -1);
                        break;
                    }
                    tempAddress -= 4;
                }
                break;
            }
                // RET.
            case 18:
            {
                shouldQuitFile = true;
                break;
            }
                // RAND.
            case 19:
            {
                short tempValue = getArgumentInteger(1);
                short tempResult = generateRandomNumber() % tempValue;
                short tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // STR.
            case 20:
            {
                short tempValue = getArgumentInteger(1);
                byte tempBuffer[20];
                sprintf((char *)tempBuffer, "%d", tempValue);
                short tempPointer = allocateText(tempBuffer);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // INT.
            case 21:
            {
                byte tempBuffer[40];
                short tempPointer = getArgumentPointer(1);
                getTextFromHeapEntry(tempBuffer, tempPointer);
                short tempResult = atoi((char *)tempBuffer);
                tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // LEN.
            case 22:
            {
                short tempResult = 0;
                short tempPointer = getArgumentPointer(1);
                while (true)
                {
                    short tempPointer2 = getHeapEntryData(tempPointer);
                    if (tempPointer2 == 0)
                    {
                        break;
                    }
                    tempResult += 1;
                    tempPointer = getHeapEntryLink(tempPointer);
                    if (tempPointer == 0)
                    {
                        break;
                    }
                }
                tempPointer = allocateInteger(tempResult);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // TRUNC.
            case 23:
            {
                short tempPointer = getArgumentPointer(0);
                short tempTargetLength = getArgumentInteger(1);
                short tempIndex = 0;
                while (tempIndex < tempTargetLength - 1)
                {
                    short tempPointer2 = getHeapEntryData(tempPointer);
                    if (tempPointer2 == 0)
                    {
                        break;
                    }
                    tempPointer2 = getHeapEntryLink(tempPointer);
                    if (tempPointer2 == 0)
                    {
                        break;
                    }
                    tempPointer = tempPointer2;
                    tempIndex += 1;
                }
                if (tempTargetLength == 0)
                {
                    setHeapEntryData(tempPointer, 0);
                }
                setHeapEntryLink(tempPointer, 0);
                break;
            }
                // GET.
            case 24:
            {
                short tempPointer = getArgumentPointer(1);
                short tempEndIndex = getArgumentInteger(2);
                short tempIndex = 0;
                while (tempIndex < tempEndIndex)
                {
                    short tempPointer2 = getHeapEntryLink(tempPointer);
                    if (tempPointer2 == 0)
                    {
                        break;
                    }
                    tempPointer = tempPointer2;
                    tempIndex += 1;
                }
                short tempPointer2 = getHeapEntryData(tempPointer);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer2);
                break;
            }
                // SET.
            case 25:
            {
                short tempPointer = getArgumentPointer(0);
                short tempEndIndex = getArgumentInteger(1);
                short tempIndex = 0;
                while (tempIndex < tempEndIndex)
                {
                    short tempPointer2 = getHeapEntryLink(tempPointer);
                    if (tempPointer2 == 0)
                    {
                        break;
                    }
                    tempPointer = tempPointer2;
                    tempIndex += 1;
                }
                while (tempIndex < tempEndIndex)
                {
                    short tempPointer2 = allocateInteger(0);
                    short tempPointer3 = allocateList(tempPointer2);
                    setHeapEntryLink(tempPointer, tempPointer3);
                    tempPointer = tempPointer3;
                    tempIndex += 1;
                }
                short tempPointer4 = getArgumentPointer(2);
                setHeapEntryData(tempPointer, tempPointer4);
                break;
            }
                // PRINT.
            case 26:
            {
                byte tempBuffer[50];
                short tempPointer = getArgumentPointer(0);
                getTextFromHeapEntry(tempBuffer, tempPointer);
                displayText(tempBuffer);
                byte tempButtons = promptButton();
                if (tempButtons & ESCAPE_BUTTON_MASK)
                {
                    hasStoppedExecution = true;
                }
                break;
            }
                // INPUT.
            case 27:
            {
                byte tempBuffer[100];
                tempBuffer[0] = 0;
                editTextLine(tempBuffer);
                short tempPointer = allocateText(tempBuffer);
                setHeapEntryReference(argumentPointerAddressList[0], tempPointer);
                break;
            }
                // Custom function.
            case 255:
            {
                byte tempFileIndex = 0;
                while (tempFileIndex < NUMBER_OF_FILE_ENTRY_POSITIONS)
                {
                    byte tempBuffer[MAXIMUM_FILE_NAME_LENGTH + 1];
                    getFileName(tempBuffer, tempFileIndex);
                    if (tempBuffer[0] != EMPTY_FILE_ENTRY_INDICATOR)
                    {
                        if (equalText(tempBuffer, tempCommandName))
                        {
                            short tempNextScopeAddress = scopeAddress + readSramShort(scopeAddress + SCOPE_SIZE_OFFSET);
                            writeSramLong(tempNextScopeAddress + SCOPE_RETURN_ADDRESS_OFFSET, tempNextCommandAddress);
                            writeSramShort(tempNextScopeAddress + SCOPE_PREVIOUS_SCOPE_ADDRESS_OFFSET, scopeAddress);
                            writeSramShort(tempNextScopeAddress + SCOPE_SIZE_OFFSET, SCOPE_FLOW_DATA_OFFSET);
                            scopeAddress = tempNextScopeAddress;
                            initializeScopeVariables();
                            byte tempIndex = 0;
                            while (tempIndex < tempNumberOfArguments)
                            {
                                short tempPointer = getArgumentPointer(tempIndex);
                                setHeapEntryReference(tempNextScopeAddress + SCOPE_VARIABLE_LIST_OFFSET + tempIndex * 2, tempPointer);
                                tempIndex += 1;
                            }
                            tempNextCommandAddress = tempFileIndex * (long)FILE_ENTRY_SIZE + FILE_DATA_OFFSET;
                            break;
                        }
                    }
                    tempFileIndex += 1;
                }
            }
            default:
            {
                break;
            }
        }
    }
    if (shouldQuitFile)
    {
        if (scopeAddress <= STACK_OFFSET)
        {
            hasStoppedExecution = true;
        } else {
            byte tempIndex = 0;
            while (tempIndex < NUMBER_OF_SCOPE_VARIABLES)
            {
                setHeapEntryReference(scopeAddress + SCOPE_VARIABLE_LIST_OFFSET + tempIndex * 2, 0);
                tempIndex += 1;
            }
            tempNextCommandAddress = readSramLong(scopeAddress + SCOPE_RETURN_ADDRESS_OFFSET);
            scopeAddress = readSramShort(scopeAddress + SCOPE_PREVIOUS_SCOPE_ADDRESS_OFFSET);
        }
    }
    commandAddress = tempNextCommandAddress;
    tempOffset = 0;
    while (tempOffset < LITERAL_ARGUMENT_ADDRESS_LIST_SIZE)
    {
        short tempAddress = LITERAL_ARGUMENT_ADDRESS_LIST_OFFSET + tempOffset;
        setHeapEntryReference(tempAddress, 0);
        tempOffset += 2;
    }
}

void displayFileMenu(byte fileIndex)
{
    long tempFileAddress = fileIndex * (long)FILE_ENTRY_SIZE;
    while (true)
    {
        byte tempResult = promptProgmemSelection(SELECTION_MENU_4, 4);
        // Edit.
        if (tempResult == 0)
        {
            editFile(fileIndex);
        }
        // Run.
        if (tempResult == 1)
        {
            scopeAddress = STACK_OFFSET;
            writeSramShort(scopeAddress + SCOPE_SIZE_OFFSET, SCOPE_FLOW_DATA_OFFSET);
            initializeScopeVariables();
            heapSize = 0;
            isIgnoringCommands = false;
            commandAddress = tempFileAddress + FILE_DATA_OFFSET;
            hasStoppedExecution = false;
            while (!hasStoppedExecution)
            {
                executeNextCommand();
            }
        }
        // Delete.
        if (tempResult == 2)
        {
            displayProgmemText(MESSAGE_4);
            promptButton();
            byte tempResult = promptProgmemSelection(SELECTION_MENU_5, 2);
            // Yes delete.
            if (tempResult == 1)
            {
                byte tempValue = EMPTY_FILE_ENTRY_INDICATOR;
                writeEepromData(tempFileAddress, &tempValue, 1);
                displayProgmemText(MESSAGE_5);
                promptButton();
                return;
            }
        }
        // Rename.
        if (tempResult == 3)
        {
            displayProgmemText(MESSAGE_2);
            promptButton();
            byte tempBuffer[50];
            tempBuffer[0] = 0;
            editTextLine(tempBuffer);
            writeEepromData(tempFileAddress, tempBuffer, MAXIMUM_FILE_NAME_LENGTH + 1);
            displayProgmemText(MESSAGE_6);
            promptButton();
        }
        if (tempResult == 255)
        {
            return;
        }
    }
}

void navigateFiles()
{
    short tempFileNumber = 0;
    while (true)
    {
        byte tempNumberOfFiles = getNumberOfFiles();
        if (tempNumberOfFiles < 1)
        {
            // No files.
            displayProgmemText(MESSAGE_8);
            promptButton();
            return;
        }
        byte tempFileIndex = getFileIndexByNumber(tempFileNumber);
        byte tempBuffer[MAXIMUM_FILE_NAME_LENGTH + 1];
        getFileName(tempBuffer, tempFileIndex);
        displayText(tempBuffer);
        byte tempButtons = promptButton();
        if (tempButtons & LEFT_BUTTON_MASK)
        {
            tempFileNumber -= 1;
        }
        if (tempButtons & RIGHT_BUTTON_MASK)
        {
            tempFileNumber += 1;
        }
        if (tempButtons & UP_BUTTON_MASK)
        {
            tempFileNumber -= 8;
        }
        if (tempButtons & DOWN_BUTTON_MASK)
        {
            tempFileNumber += 8;
        }
        if (tempFileNumber < 0)
        {
            tempFileNumber = 0;
        }
        if (tempFileNumber >= tempNumberOfFiles)
        {
            tempFileNumber = tempNumberOfFiles - 1;
        }
        if (tempButtons & RETURN_BUTTON_MASK)
        {
            displayFileMenu(tempFileIndex);
            tempFileNumber = 0;
        }
        if (tempButtons & ESCAPE_BUTTON_MASK)
        {
            return;
        }
    }
}

void displayPrimaryMenu()
{
    while (true)
    {
        byte tempResult = promptProgmemSelection(SELECTION_MENU_3, 2);
        // Create.
        if (tempResult == 0)
        {
            displayProgmemText(MESSAGE_2);
            if (promptButton() & ESCAPE_BUTTON_MASK)
            {
                continue;
            }
            byte tempBuffer[50];
            tempBuffer[0] = 0;
            editTextLine(tempBuffer);
            tempBuffer[FILE_DATA_OFFSET] = 0;
            long tempFileAddress = 0;
            while (true)
            {
                byte tempValue = readEepromByte(tempFileAddress);
                if (tempValue == EMPTY_FILE_ENTRY_INDICATOR)
                {
                    break;
                }
                tempFileAddress += FILE_ENTRY_SIZE;
            }
            writeEepromData(tempFileAddress, tempBuffer, FILE_DATA_OFFSET + 1);
            displayProgmemText(MESSAGE_3);
            promptButton();
        }
        // Select.
        if (tempResult == 1)
        {
            navigateFiles();
        }
        // Sure we have room for an easter egg.
        if (tempResult == 255)
        {
            displayProgmemText(MESSAGE_9);
            promptButton();
        }
    }
}

int main(int argc, const char * argv[])
{
    std::cout << "Starting..." << std::endl;
    
    long tempIndex = 0;
    while (tempIndex < sizeof(eepromBuffer))
    {
        eepromBuffer[tempIndex] = EMPTY_FILE_ENTRY_INDICATOR;
        if (tempIndex % 10000 == 0)
        {
            std::cout << tempIndex << std::endl;
        }
        tempIndex += 1;
    }
    
    byte tempFileIndex = 0;
    byte tempArgumentIndex = 1;
    while (tempArgumentIndex < argc)
    {
        byte tempFileName[] = "MYFILE ";
        tempFileName[6] = tempFileIndex + '0';
        tempIndex = 0;
        while (tempIndex < sizeof(tempFileName))
        {
            byte tempCharacter = tempFileName[tempIndex];
            eepromBuffer[tempFileIndex * FILE_ENTRY_SIZE + tempIndex] = tempCharacter;
            tempIndex += 1;
        }
        std::ifstream tempFile;
        std::cout << argv[tempArgumentIndex] << std::endl;
        tempFile.open(argv[tempArgumentIndex]);
        if (tempFile.is_open()) {
            tempFile.seekg(0, std::ios::end);
            std::streamsize tempAmount = tempFile.tellg();
            tempFile.seekg(0, std::ios::beg);
            long tempAddress = tempFileIndex * FILE_ENTRY_SIZE + FILE_DATA_OFFSET;
            tempFile.read((char *)(eepromBuffer + tempAddress), tempAmount);
            eepromBuffer[tempAddress + tempAmount] = '\n';
            eepromBuffer[tempAddress + tempAmount + 1] = 0;
        }
        tempFile.close();
        tempFileIndex += 1;
        tempArgumentIndex += 1;
    }
    
    displayProgmemText(MESSAGE_7);
    promptButton();
    
    randomNumber = time(NULL);
    
    displayPrimaryMenu();
    
    return 0;
}

